#include <iostream>
#include <vector>
#include <queue>
#include <thread>
#include <mutex>
#include <condition_variable>
#include <functional>
class ThreadPool {
public:
    ThreadPool(int numThreads) : stop(false) {
        //iterates numThreads times, creating a worker thread for each iteration
        for (int i = 0; i < numThreads; ++i) {  
            // Create a worker thread and add it to the workers vector
            workers.emplace_back([this, i] {
                while (true) {
                    std::function<void()> task;
                    {
                        std::unique_lock<std::mutex> lock(queueMutex);
                        // Wait until there is a task available or the stop flag is set
                        condition.wait(lock, [this] { return stop || !tasks.empty(); });
                        // If the stop flag is set and the task queue is empty, exit the thread
                        if (stop && tasks.empty())
                            return;
                        // Move the front task from the queue to the local task variable
                        task = std::move(tasks.front());
                        tasks.pop();
                    }
                    // Execute the task
                    task();
                }
            });
        }
    }
    ~ThreadPool() {
        {
            std::unique_lock<std::mutex> lock(queueMutex);
            stop = true;
        }
        condition.notify_all();
        for (std::thread& worker : workers)
            worker.join();
    }
    template<typename F, typename... Args>
    //A function func with args... as input are passed as rvalue references
    void enqueue(F&& func, Args&&... args) { 
        {
            std::unique_lock<std::mutex> lock(queueMutex);
            tasks.emplace([func, args...] { func(args...); }); // a task to be executed
        }
        condition.notify_one(); //notifies one worker thread
    }
private:
    std::vector<std::thread> workers;
    std::queue<std::function<void()>> tasks;
    std::mutex queueMutex;
    std::condition_variable condition;
    bool stop;
};
// Example of a computation
void processTask(int taskId) {
    std::cout << "Processing task " << taskId << " in thread " << std::this_thread::get_id() << std::endl;
    std::this_thread::sleep_for (std::chrono::seconds(1)); // task processing
}
int main() {
    const int numTasks = 10;
    const int numThreads = 3 ;//std::thread::hardware_concurrency(); 
    std::cout << "Executing " << numTasks << " tasks in a thread pool of: " << numThreads << " threads" << std::endl;
    ThreadPool threadPool(numThreads);
    for (int i = 0; i < numTasks; ++i) {
        threadPool.enqueue(processTask, i);
    }
    // Wait for all tasks to complete
    std::this_thread::sleep_for(std::chrono::seconds(1));
    return 0;
}
